-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
--[[

    Responsibile for loading all the maps that the server requires to be loaded
    
--]]

require('strict')
require('utils')
class = require('middleclass')

local PickableManager = require('pickable_manager')

local ServerMapLoader = class('ServerMapLoader')

require('objects/buried_item_manager')
require('procedural_generation/procedural_generation')

gulp.maps = {}
gulp.loadable_maps = {}

map_loading_buffer = {}

-- helper function for map loading
function RawMapData(m)
    if type(m) == "table" then
                
        map_loading_buffer = m
        
    end
end

function ServerMapLoader:load_map_file(name)
    
    -- In future, index maps by map name, and only load the map from file if it is not already in the table??
    -- we'll need to do network load of maps I guess, in case the same maps have been edited differently on 
    -- server and client. We should just download the whole map file to the clients when the server gets here. 
 
    local success, err = pcall(dofile, gulp_cpp.LoadPath("maps/" .. name .. ".lua"):str())
    if not success then
        print("Pcall Error - Loading Maze", name, err)
    else
        local map = map_loading_buffer
        
        if map.procedural ~= nil then
            -- do some procedural generation!
            procedurally_generate(map)
        end
        
        map.auto = {}
        map.auto.filename = name
        
        map.auto.wall_detect_type = {}
        map.auto.glyph_mappings = {}
        map.auto.autowall_like = {}
        map.auto.autowall = {}
        map.auto.wall_glyph_lookup_table_transformed = {}
        
        map = tabulate_map(map)
        
        -- save this map into a loadable buffer before we mess with it any further
        -- this is for passing to clients
        -- do a deep copy so that future changes to map are not reflected in loadable_map[map_name]
        gulp.loadable_maps[name] = deep_copy(map)
        
        
        map.auto.maze = gulp_cpp.PresentationMaze()
        map.auto.mobs = {}
        map.auto.pickables = PickableManager(name)          -- pickable manager needs to know who it belongs to
        map.auto.objects = {}
        
        -- set a default game manager to be loaded
        if map.manager_name == nil then map.manager_name = 'default_game_manager' end
        if map.manager_file == nil then map.manager_file = map.manager_name end

        
        if map.auto.maze == nil or map.auto.maze == 0 then
            print("FATAL - no maze allocated in load_map_file")
            error("FATAL - no maze allocated in load_map_file")
        else
            gulp.maps[name] = map
        end

    end


end



function ServerMapLoader:initialize(map_name)
    
    -- garbage collector data
    self.maps_to_be_garbage_collected = 
    {
        'coin_rush',
        'coin_rush_lobby',
        'maze_tag_lobby',
        'maze_tag_2_player',
        'maze_tag_2_player_b',
        'maze_tag_2_player_c',
        'maze_tag_3_player',
        'maze_tag_3_player_b',
        'maze_tag_3_player_c',
        'maze_tag_4_player',
        'maze_tag_4_player_b',
        'maze_tag_4_player_c',
        'maze_tag_4_player_d',
    }
    
    -- to start with, all maps are pristine
    self.pristine = {}
    for _, map in ipairs(self.maps_to_be_garbage_collected) do
        self.pristine[map] = true
    end
    


end

function ServerMapLoader:DropAllMaps()
    for name,map in pairs(gulp.maps) do
        self:DropMap(name)
    end
end


function ServerMapLoader:DropMap(map_name)
    
    if not check_map_exists(map_name) then return end
    local map = get_map(map_name)
    
    
    for _, object in ipairs(map.auto.objects) do
        object:finish()
    end
    
    for _, mob in pairs(map.auto.mobs) do
        mob:finish()
    end
    
    map.auto.mobs = {}
    map.auto.objects = {}
    map.auto.pickables:delete_all()
    
    local manager_name = map.manager_name
    gulp.maps[map_name] = nil
    
    -- must drop the game manager after the map is deleted,
    -- or it will be seen as still in use
    drop_game_manager(manager_name)
end

function ServerMapLoader:SwitchMaze(player_id, map_name)
    -- check if the required map already exists,
    -- if so place player in that map, otherwise it should load the map
    -- into a new entry in gulp.maps
    
    -- if we don't hold this maze in memory, load it
    if not check_map_exists(map_name) then
        self:LoadMap(map_name)
    end
   
    
    local source_player = get_player_by_id(player_id)
    if source_player then 
        
        -- switch the player to the correct maze
        source_player.map_name = map_name
        
        -- broadcast a player position message to let all clients know who is on what map
        source_player:broadcast_position('switch_map') 
    end
    
    return true
    
end

function ServerMapLoader:LoadMap(map_name)
    gulp.maps[map_name] = nil
    self:load_map_file(map_name)
    load_maze_assets(map_name)
    generate_treasure(map_name)
end

function ServerMapLoader:form_map_message(map_name)
    
    if not check_map_exists(map_name) then
        self:LoadMap(map_name)
    end
    
    local message = {
        mt = 'map_file',
        map_file_name = map_name,
        file = gulp.loadable_maps[map_name],
    }
    
    return message
end

    

function ServerMapLoader:update(dt)
    self:map_garbage_collect()
end


function ServerMapLoader:map_garbage_collect()
    -- Many maps should be persistent across the game, but some shouldn't
    -- This function checks for non-persistent maps which have no players on them, and deletes them
    
    if gulp.LuaStateCodeMode == 'SERVER' then
        
        -- get a list of all the maps currently being used
        local map_has_player = {}
        for _,player in all_players() do
            map_has_player[player.map_name] = true
        end
        
        -- for every map which we might garbage collect...
        for _,map in pairs(self.maps_to_be_garbage_collected) do
            
            -- only if we have it loaded...
            if check_map_exists(map) then
            
                -- if there is a player on it, it is no longer pristine, but
                -- since there is a player, we do not garbage collect it 
                if map_has_player[map] == true then
                    self.pristine[map] = false
                else
                    
                    -- there is no player, so garbage collect it unless it is already pristine
                    if not self.pristine[map] then
                        --print("Garbage collecting "..map)
                        
                        -- drop the map and mark it as pristine for the next time it is loaded
                        self:DropMap(map)
                        self.pristine[map] = true
                
                        -- tell clients to drop the map too
                        for _,player in all_players() do
                            if player.loaded_maps~=nil and player.loaded_maps[map] == true then
                                local message = {
                                        mt = 'reset_map',
                                        map_name = map,
                                    }
                                player.socket_ref:SendMessage(message)
                            end
                        end
                    end
                end
            end
        end
    end
end

        
        
        


return ServerMapLoader