-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

require('strict')
require('utils')


function load_game_manager(map_name)
    
    local map = get_map(map_name)
    local manager_name = map.manager_name
    local manager_file = map.manager_file
    
    -- if the required game controller does not exist, create it
    if gulp.game_managers[manager_name] == nil then
        local gm = require("map_controllers/" .. manager_file)
        gulp.game_managers[manager_name] = gm:new()
        gulp.game_managers[manager_name].my_name = manager_name
        gulp.game_managers[manager_name].map_name = map_name
        gulp.game_managers[manager_name]:post_map_initialisation()
    end

end


function find_game_manager_by_name(manager_name)
    
    if manager_name == nil then
        error("FATAL: map.manager_name is not specified")
    end
    
    local manager = gulp.game_managers[manager_name]
    
    if manager == nil then
        error("FATAL: No game manager loaded for map")
    end
    
    return manager
    
end

function find_game_manager(map)
    return find_game_manager_by_name(map.manager_name)
end


function find_local_game_manager()
    return find_game_manager(get_map(gulp.localplayer.map_name))
end

function find_player_game_manager(id)
    if id == 0 then 
        return find_local_game_manager() 
    else
        local rp = get_player_by_id(id)
        local map = get_map(rp.map_name)
        return find_game_manager(map)
    end
end

function drop_game_manager(manager_name)
    
    local manager = gulp.game_managers[manager_name]
    
    -- first check if the manager is still in use
    if next(gulp.maps) ~= nil then
        for k,map in pairs(gulp.maps) do
            if map.manager_name == manager_name then
                -- this manager is still in use
                return
            end
        end
    end
    
    -- then check the manager exists
    if manager == nil then return end
    
    -- if exists and not in use, then delete it
    manager:finish()
    gulp.game_managers[manager_name] = nil
    
end



local class = require('middleclass')


local GameManager = class('GameManager')

function GameManager:initialize()
end

function GameManager:finish()
end


--[[
-- draw() and update() don't need to be provided, since 
-- the dispatcher checks for their existence

function GameManager:draw()
end

function GameManager:update(dt)
end

--]]
function GameManager:post_map_initialisation()
    -- must be provided but does nothing by default
    -- called after construction of the game manager, once the map name has been set
end

function GameManager:send_game_manager_status(socket_ref)
    -- must be provided but does nothing by default
end


function GameManager:notify(player_id, action)
    -- must be provided but does nothing by default
end

function GameManager:map_conditional_notify(command, conditional, value, new_glyph)
    -- checks doors - if you override and want doors to work, remember to call this too
    local map = get_map(self.map_name)
    local maze = map.auto.maze
    
    for _, door in pairs(map.doors) do
        if door.conditional == conditional then
            set_map_glyph(self.map_name, door.line+1, door.column+1, door[value])
        end
    end
        
end


function GameManager:process_game_control_message(message)
    -- must be provided but does nothing by default
end             

function GameManager:switch_maze(player_id, new_maze_name)
    -- must be provided, by default just passes action to gulp.map_loader
    return gulp.map_loader:SwitchMaze(player_id, new_maze_name)
end

function GameManager:cause_drawing()
    
    local localplayer = gulp.localplayer
    
    -- we need this check to cope with startup and teardown
    if localplayer ~= nil then
        
        -- we need this check to cope with maze switching which may have 
        -- discarded a map before the player's map has been updated
        if check_map_exists(localplayer.map_name) then
            local map = get_map(localplayer.map_name)
    
            -- only draw if we are managing the current map
            if self.my_name == map.manager_name then
                self.need_to_draw = true
            end
        end
    end
end

        

return GameManager
