-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

require("strict")
local class = require("middleclass")
local copas = require("copas")

-- Message queue for copas sending 
local send_message_queues = {}

local sockets_to_close_list = {}

function MessagesInClientQueue(socket_ref)
    return send_message_queues[socket_ref]
end



local NetworkTransport = class("NetworkTransport")

function NetworkTransport:initialize(connection_socket)
    self.my_socket = connection_socket
end

-- wrapper to add a message to the queue for sending   
function NetworkTransport:SendMessage(message)
    
    -- temporary until everthing has been converted over
    if type(message) == "string" then
        message = { mt = "old", content = message }
    end 
    local message_string = encode_for_network(message)

    local raw_socket = self.my_socket
    -- print(send_message_queues[nil]) prints nil
    -- if send_message_queues[socket_ref] then is always false!
    if raw_socket == nil then
        print("Socket is nil - dropping message")
        --local traceback = debug.traceback()
        --for line in traceback:gmatch("(.-)\r?\n") do
        --    print("   >>>", line)
        --end
    elseif send_message_queues[self] then
        -- the queue for that message exists
        table.insert(send_message_queues[self], message_string)
    else
        -- doesn't exist ... create it
        send_message_queues[self] = { message_string }
    end
end

function NetworkTransport:close()
    self.my_socket:close()
end

function NetworkTransport:receive(pattern, part)
    copas.recieve(self.my_socket, pattern, part)
end

-- Add the socket to be closed to the list 
function GulpNetworking.AddSocketToClose(socket)
    -- Add the socket the queue
    table.insert(sockets_to_close_list, socket)
end

-- Copas thread that will pop the message and send over the network    
function GulpNetworking.SendMessageThread()
  while true do
    for socket_ref, queue in pairs(send_message_queues) do
        -- if the socket looks to be be something then try and send
        local messages = table.concat(queue)
        -- remove queue, we've sent it
        send_message_queues[socket_ref] = nil
        
        -- This might yield if send doesn't go.
        -- It also yields 1 in 10 (on average, actually at random).
        -- This affects ALL client transmissions not just this one.
        -- We could use a seperate thread per client, but since this is the
        -- server sending (just to different sockets), does that help?
        -- Yielding 1 in 10 means messages are delayed by one frame, 
        -- potentially.
        local s, err, lastIndex = copas.send(socket_ref.my_socket, messages)
        if (err) then
            -- could be 'closed' or 'timeout' - but 
            -- copas uses timeout, so must be close!
            print ("SendMessageThread - send error on on socket: ", socket_ref)
            print ("The Copas error is: ", err)
         
            if gulp.LuaStateCodeMode == 'SERVER' then
                GulpNetworking.HandleNetworkErrorServer(socket_ref, "SendError")
            else
                GulpNetworking.HandleNetworkErrorClient(socket_ref, "SendError")
            end
        else
            -- @todo: debug code - remove after testing
            if s ~= #messages then
                print("SendMessageThread - Didn't send all of message" , s, #messages)
            end
        end
    end
    
    if #sockets_to_close_list > 0 then
        copas.sleep(0.1)

        -- close all sockets on the close list
        for index, socket in ipairs(sockets_to_close_list) do
            socket:close()
        end

        sockets_to_close_list = {}
        print(gulp.LuaStateCodeMode, "All Sockets closed - return from SendMessageThread")
        return
    end
        
    copas.sleep(0) -- use 0 just to other copas threads to run  
  end
end  


function GulpNetworking:stepCopas()
    -- No delay
    copas.step(0)
end


function GulpNetworking.HandleNetworkErrorServer(socketWithError, ErrorType)
    print ("Server Network Error: ", ErrorType)
    
    -- Handle any serve stuff
    -- Remove the remote player as there connection has failed
    HandleClientUnregisterMessage(socketWithError, "dummy as not used")  
        
    if send_message_queues[socketWithError] then
        -- Error with this socket so clear to nil to prevent us trying to use it again
        send_message_queues[socketWithError] = nil -- make sure nothing left on old socket number (copas.send might switch)
    end
end

function GulpNetworking.HandleNetworkErrorClient(socketWithError, ErrorType)
    print ("Client Network Error: ", ErrorType)
    
    -- Tell the rest of the game that an error has occured.
    gulp.setNetworkError()
    
    if send_message_queues[socketWithError] then
        -- Error with this socket so clear to nil to prevent us trying to use it again
        send_message_queues[socketWithError] = nil -- make sure nothing left on old socket number (copas.send might switch)
    end
end


return NetworkTransport
