-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Server Remote Player
-- Used on server side only
-- Handles stuff that needs to be done on the server for the remote players
--
-- * map changes
-- * pickable objects

-- need objects if they are to be droppable


require('strict')
require('utils')
class = require('middleclass')
avatars = require('avatars/avatar_loader')

local PlayerBase = require("player/PlayerBase")

local ServerRemotePlayer = class('ServerRemotePlayer', PlayerBase)

local GoldCoin = require('objects/gold_coin')
local Hole = require('objects/hole')

-- Creates the player
function ServerRemotePlayer:initialize(socket_ref, message, id)
    -- Store the player name and id
    self.socket_ref = socket_ref
  
    -- decode register message
    self.playerName= message.name 
    
    -- initialise the subclass
    PlayerBase.initialize(self, message.avatar, message.colour, id, message.map_name, message.line, message.column)
    
    gulp_networking:SendMobsAndPickables(socket_ref, self.map_name)
    gulp_networking:SendPlayersAndGameStatus(self, self.map_name)
    
    -- add a delegate for detector mode (maybe other stuff too)
    gulp._delegate_list:add_delegate(self)
    self.ping_accumulator = 0

    
    -- used to route messages and inform the server's map garbage collector
    -- assume the client already has the starting map
    self.loaded_maps = { [gulp.server_initial_map_name] = true }
  
end

function ServerRemotePlayer:die()
    gulp_networking:SendDeath(self.socket_ref)
end

-- Remove the user 
function ServerRemotePlayer:remove()
  
    gulp_networking:SendServerGameEnd(self.socket_ref)
  
    PlayerBase.remove(self)
  
    -- Close the network socket
    if self.socket_ref then
        --self.socket_ref:close()
    end 
    
    self.id = nil
end

function ServerRemotePlayer:take_exit(exit_to_take)
    local message = {
        mt = 'player_control',
        command = 'take_exit',
        exit = exit_to_take,
    }
    self.socket_ref:SendMessage(message)
end    
  
function ServerRemotePlayer:register_map(map_name)
    self.loaded_maps[map_name] = true
    
    --debug
    --print(self.id..": register_map "..map_name)
end

function ServerRemotePlayer:drop_map(map_name)
    self.loaded_maps[map_name] = nil
    
    --debug
    --print(self.id..": drop_map "..map_name)
end
    
  
  
-- Updates the player postion on the maze   
function ServerRemotePlayer:processPosition(message)
    
    local dl = message.line - self.line
    local dc = message.column - self.column
    
    local collision 
    collision, dl, dc = get_collision(self.collision_ref, message.map_name, self.line, self.column, dl, dc) 
    
    if collision then
        -- update client on the real player position - adjusted for collision
        local message = {
            mt = 'player_control',
            command = 'collision',
            line = self.line+dl,
            column = self.column+dc,
        }
        self.socket_ref:SendMessage(message)            
    end

    
    self.line = self.line+dl
    self.column = self.column+dc
    
    local player_map = message.map_name
    
    if self.id then
        
        -- TEMPORARY
        -- Player has moved, now they are fair game...
        -- It used to make the player visible on map change, which happened at startup, but now no longer does
        -- we need a better solution here
        self.visible_to_mobs = true
    
        -- check for moving maps
        if self.map_name ~= player_map then
            -- we have moved maps (or else this is startup)
            -- as a server, I have a responsibility to provide the client with mobs and pickables from the new map
                
            find_player_game_manager(self.id):switch_maze(self.id, player_map)
            gulp_networking:SendPlayersAndGameStatus(self, player_map)
            
        end
        
        -- check for death!
        
        -- check for pickables
        local item = get_map(self.map_name).auto.pickables:pick_up_first(self.line, self.column)
        if item then
            
            if item.name == 'gold coin' then
                find_player_game_manager(self.id):notify(self.id, 'pick_up_coin')
            end
            
            gulp_networking:SendPickedItem(self.socket_ref, item)
        end
        
        -- check for floor effects
        -- currently we use avatar centre, could start using the edges like wall detection, 
        -- but then we might be applying up to 4 effects!
        local floor_effect = is_active_floor(self.map_name, self.line, self.column)
        if floor_effect ~= nil then
            
            -- player is stood on an active floor - do the floor activity
            if floor_effect[1] == 'die' then
                self:die()
            end
        end
        
        
    end
    
    PlayerBase.processPosition(self, message)

end

local handled_object_types = {
    ['gold_coin'] = { GoldCoin, "gold coin", function(player) 
            find_player_game_manager(player.id):notify(player.id, 'drop_coin') 
            end },
    ['hole'] =      { Hole,     "hole",      function(player, line, column) 
            get_map(player.map_name).auto.pickables:check_for_buried_treasure(line, column)
            end },
    }

function ServerRemotePlayer:processDroppedItem(line, column, item_type) 
    local table_entry = handled_object_types[item_type]
    if table_entry then
        local item = table_entry[1](table_entry[2])
        get_map(self.map_name).auto.pickables:add_to_floor(line, column, item)
        table_entry[3](self, line, column)
    else
        print("Tried to network drop unknown object " .. item_type)
    end
end

-- Process the player action
function ServerRemotePlayer:processAction(player_action)
    --print("PlayerBase:processAction() - action:", player_action)
    if player_action == 'detect_mode' then
        self.detect_mode = true
    elseif player_action == 'detect_mode_cancel' then
        self.detect_mode = false
    end
end

function ServerRemotePlayer:update(dt)
    if self.detect_mode == true then
        self.ping_accumulator = self.ping_accumulator + dt
    
        if self.ping_accumulator > 0.20 then
            self.ping_accumulator = self.ping_accumulator - 0.20
            
            -- get the ping distance
            local dist = get_map(self.map_name).auto.pickables:get_distance_to_metallic_treasure(self.line, self.column, 3)
            
            local message = {
                mt = 'player_control',
                command = 'detector',
                distance = dist,
            }
            self.socket_ref:SendMessage(message)            
        end
    end
end

return ServerRemotePlayer
