-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require('strict')

function print_indexed_table(table_to_print, indent)
    local indent = indent or 0
    for k, v in ipairs(table_to_print) do
        local formatting = string.rep("  ", indent) .. "[" .. k .. "] = "
        if type(v) == "table" then
            print(formatting)
            print_indexed_table(v, indent+1)
        elseif type(v) == 'boolean' or type(v) == 'function' then
            print(formatting .. tostring(v))      
        else
            print(formatting .. v)
        end
    end
end

function print_table(table_to_print, indent)
    local indent = indent or 0
    for k, v in pairs(table_to_print) do
        
        local formatting
        if type(k) == 'table' then
            formatting = string.rep("  ", indent) .. "[table]" .. ": "
        else
            local key_ = k
            if type(k) == "boolean" then
                key_ = tostring(k)
            end
            formatting = string.rep("  ", indent) .. key_ .. ": "
        end
        
        if string.find(formatting, "__") == 1 then
            ;-- skip this internal item
        else
               
            if type(v) == "table" then
                print(formatting)
                if indent > 20 then
                    print("tables nested 20 deep?")
                else
                    print_table(v, indent+1)
                end
            elseif type(v) == 'boolean' or type(v) == 'function' or type (v) == 'userdata' then
                print(formatting .. tostring(v))      
            else
                print(formatting .. v)
            end
        end
        
    end
end

function round(float)
    if float > 0 then
        return math.floor(float+0.5)
    else
        return math.ceil(float-0.5)
    end
end


function show_versions()
    print("Egg Tag")
    print("Version:", gulp.forlorn_fox_version)
    print("Engine:", gulp.forlorn_fox_engine_version)
    local compiled = gulp_cpp.SDL_version()
    gulp_cpp.SDL_VERSION(compiled)
    print("SDL Compiled against:", compiled.major, compiled.minor, compiled.patch)
    local linked = gulp_cpp.SDL_version()
    gulp_cpp.SDL_GetVersion(linked)
    print("SDL Running against:", linked.major, linked.minor, linked.patch)
    print("SDL Revision", gulp_cpp.SDL_GetRevision())
    print("SDL Revision Num", gulp_cpp.SDL_GetRevisionNumber())
    print("Interpreter:", _VERSION)
end



-- find things close to other things
-- 'me' is a table that must contain line, column and map_name
-- 'target' is a table of tables, which must each contain line, column and map_name
-- otherwise, you can check whether anything is close to anything else
-- 'iterator' controls how the list is stepped over, see PlayerList.lua for funky examples, 
--   or otherwise just leave it off the call for default operation
function find_nearby(me, targets, nearby_distance, iterator, mode)
    
    local nearby_keys = {}
    
    if not iterator then iterator = pairs end
    if not mode then mode = 'quick' end
    
    for key,target in iterator(targets) do
        
        local line_delta = math.abs(me.line - target.line)        
        local column_delta = math.abs(me.column - target.column)
        
        if  me.map_name == target.map_name then
            if mode == 'quick' then
                if line_delta < nearby_distance and column_delta < nearby_distance then
                    -- we are nearby!
                    table.insert(nearby_keys, key)
                end
                
            else 
                -- do proper distance calcs
                local d = math.sqrt((line_delta*line_delta) + (column_delta*column_delta))
            
                if d <= nearby_distance then
                    -- we are nearby
                    if mode == 'return_distances' then
                        table.insert(nearby_keys, {key, d})
                    else
                        table.insert(nearby_keys, key)
                    end
                end
    
            end
        end
    end
    
    return nearby_keys
end

function get_nearby_distances(me, targets, nearby_distance, iterator)
    if not iterator then iterator = pairs end
    return find_nearby(me, targets, nearby_distance, iterator, 'return_distances')
end


function random_entry(itable)
    return itable[math.random(#itable)]
end

function table_is_empty(table)
    return type(table)=='table' and next(table) == nil
end

function check_map_exists(map_name)
    return gulp.maps[map_name] ~= nil
end

function get_map(map_name)
    if gulp.maps[map_name] == nil then
        error("Map "..map_name.." not found")
    else
        return gulp.maps[map_name]
    end
end

function get_maze(map_name)
    local map = get_map(map_name)
    return map.auto.maze
end

-- basic deep copy
-- does not handle metatables 
-- does not handle recursive tables - class tables are recursive
function deep_copy(item)
    if type(item) ~= 'table' then 
        return item 
    end
    
    local copy = {}
    for k,v in pairs(item) do
        copy[k] = deep_copy(v)
    end
    
    return copy
end

function percent_chance(percentage)
    return math.random(100) < percentage
end

function time_function(message, fn, arg1, arg2, arg3, arg4, arg5)
    local start_t = gulp_cpp.SDL_GetTicks()
    local ret_val = fn(arg1, arg2, arg3, arg4, arg5)
    local time = gulp_cpp.SDL_GetTicks() - start_t
    
    if gulp.DebugLevelLoading then
        print(message, ":", time)
    end
    
    return ret_val
end

function gulp.debug_print(key, display_type, line, column, param1, param2, param3)
    if not gulp.server_debug_print_table then
        gulp.server_debug_print_table = {}
    end
    if display_type then
        gulp.server_debug_print_table[key] = { display_type, line, column, param1, param2, param3 }
    else
        gulp.server_debug_print_table[key] = nil
    end
end
    
function split(s, delimiter)
    local results = {}
    local max_width = 0
    
    for match in (s..delimiter):gmatch("(.-)"..delimiter) do
        local entry = { ['word'] = match, ['width'] = ANSI_printable_len(match) }
        table.insert(results, entry)
        if entry['width'] > max_width then max_width = entry['width'] end
    end
    return results, max_width
end    

function render_flag_decode(rend)

    local f = rend.flags

    local S = rend.software_flags_mask
    local A = rend.accleratated_flags_mask
    local V = rend.vsync_flags_mask
    local T = rend.target_texture_flags_mask

    local lookup = { S, A, V, T }
    local chars = { "S", "A", "V", "T" }
    local t = ""
    for i, k in ipairs(lookup) do
        if bit32.btest(f, k) then
            t = t .. chars[i]
            f = f - k
        end
    end
    if f ~= 0 then
        t = t .. "?"
    end
    
    return t
end

